@@ -1,12 +1,10 @@
 /**
  * @file C grammar for tree-sitter
- * @author Max Brunsfeld
+ * @author Max Brunsfeld <maxbrunsfeld@gmail.com>
+ * @author Amaan Qureshi <amaanq12@gmail.com>
  * @license MIT
  */
 
-/* eslint-disable arrow-parens */
-/* eslint-disable camelcase */
-/* eslint-disable-next-line spaced-comment */
 /// <reference types="tree-sitter-cli/dsl" />
 // @ts-check
 
@@ -22,12 +20,12 @@ const PREC = {
   BITWISE_AND: 5,
   EQUAL: 6,
   RELATIONAL: 7,
-  SIZEOF: 8,
-  OFFSETOF: 9,
-  SHIFT: 10,
-  ADD: 11,
-  MULTIPLY: 12,
-  CAST: 13,
+  OFFSETOF: 8,
+  SHIFT: 9,
+  ADD: 10,
+  MULTIPLY: 11,
+  CAST: 12,
+  SIZEOF: 13,
   UNARY: 14,
   CALL: 15,
   FIELD: 16,
@@ -60,10 +58,14 @@ module.exports = grammar({
     [$._type_specifier, $._expression_not_binary],
     [$._type_specifier, $._expression_not_binary, $.macro_type_specifier],
     [$._type_specifier, $.macro_type_specifier],
+    [$._type_specifier, $.sized_type_specifier],
     [$.sized_type_specifier],
     [$.attributed_statement],
     [$._declaration_modifiers, $.attributed_statement],
     [$.enum_specifier],
+    [$._type_specifier, $._old_style_parameter_list],
+    [$.parameter_list, $._old_style_parameter_list],
+    [$.function_declarator, $._function_declaration_declarator],
   ],
 
   word: $ => $.identifier,
@@ -74,6 +76,7 @@ module.exports = grammar({
     // Top level items are block items with the exception of the expression statement
     _top_level_item: $ => choice(
       $.function_definition,
+      alias($._old_style_function_definition, $.function_definition),
       $.linkage_specification,
       $.declaration,
       $._top_level_statement,
@@ -90,6 +93,7 @@ module.exports = grammar({
 
     _block_item: $ => choice(
       $.function_definition,
+      alias($._old_style_function_definition, $.function_definition),
       $.linkage_specification,
       $.declaration,
       $._statement,
@@ -144,6 +148,8 @@ module.exports = grammar({
 
     ...preprocIf('', $ => $._block_item),
     ...preprocIf('_in_field_declaration_list', $ => $._field_declaration_list_item),
+    ...preprocIf('_in_enumerator_list', $ => seq($.enumerator, ',')),
+    ...preprocIf('_in_enumerator_list_no_comma', $ => $.enumerator, -1),
 
     preproc_arg: _ => token(prec(-1, /\S([^/\n]|\/[^*]|\\\r?\n)*/)),
     preproc_directive: _ => /#[ \t]*[a-zA-Z0-9]\w*/,
@@ -223,14 +229,27 @@ module.exports = grammar({
     function_definition: $ => seq(
       optional($.ms_call_modifier),
       $._declaration_specifiers,
+      optional($.ms_call_modifier),
       field('declarator', $._declarator),
       field('body', $.compound_statement),
     ),
 
+    _old_style_function_definition: $ => seq(
+      optional($.ms_call_modifier),
+      $._declaration_specifiers,
+      field('declarator', alias($._old_style_function_declarator, $.function_declarator)),
+      repeat($.declaration),
+      field('body', $.compound_statement),
+    ),
+
     declaration: $ => seq(
       $._declaration_specifiers,
       commaSep1(field('declarator', choice(
-        seq($._declarator, optional($.gnu_asm_expression)),
+        seq(
+          optional($.ms_call_modifier),
+          $._declaration_declarator,
+          optional($.gnu_asm_expression),
+        ),
         $.init_declarator,
       ))),
       ';',
@@ -239,13 +258,13 @@ module.exports = grammar({
     type_definition: $ => seq(
       optional('__extension__'),
       'typedef',
-      repeat($.type_qualifier),
-      field('type', $._type_specifier),
-      repeat($.type_qualifier),
-      commaSep1(field('declarator', $._type_declarator)),
+      $._type_definition_type,
+      $._type_definition_declarators,
       repeat($.attribute_specifier),
       ';',
     ),
+    _type_definition_type: $ => seq(repeat($.type_qualifier), field('type', $._type_specifier), repeat($.type_qualifier)),
+    _type_definition_declarators: $ => commaSep1(field('declarator', $._type_declarator)),
 
     _declaration_modifiers: $ => choice(
       $.storage_class_specifier,
@@ -341,6 +360,15 @@ module.exports = grammar({
       $.identifier,
     ),
 
+    _declaration_declarator: $ => choice(
+      $.attributed_declarator,
+      $.pointer_declarator,
+      alias($._function_declaration_declarator, $.function_declarator),
+      $.array_declarator,
+      $.parenthesized_declarator,
+      $.identifier,
+    ),
+
     _field_declarator: $ => choice(
       alias($.attributed_field_declarator, $.attributed_declarator),
       alias($.pointer_field_declarator, $.pointer_declarator),
@@ -370,21 +398,25 @@ module.exports = grammar({
 
     parenthesized_declarator: $ => prec.dynamic(PREC.PAREN_DECLARATOR, seq(
       '(',
+      optional($.ms_call_modifier),
       $._declarator,
       ')',
     )),
     parenthesized_field_declarator: $ => prec.dynamic(PREC.PAREN_DECLARATOR, seq(
       '(',
+      optional($.ms_call_modifier),
       $._field_declarator,
       ')',
     )),
     parenthesized_type_declarator: $ => prec.dynamic(PREC.PAREN_DECLARATOR, seq(
       '(',
+      optional($.ms_call_modifier),
       $._type_declarator,
       ')',
     )),
     abstract_parenthesized_declarator: $ => prec(1, seq(
       '(',
+      optional($.ms_call_modifier),
       $._abstract_declarator,
       ')',
     )),
@@ -425,17 +457,32 @@ module.exports = grammar({
       field('declarator', $._type_declarator),
     ))),
     abstract_pointer_declarator: $ => prec.dynamic(1, prec.right(seq('*',
+      repeat($.ms_pointer_modifier),
       repeat($.type_qualifier),
       field('declarator', optional($._abstract_declarator)),
     ))),
 
     function_declarator: $ => prec.right(1,
+      seq(
+        field('declarator', $._declarator),
+        field('parameters', $.parameter_list),
+        optional($.gnu_asm_expression),
+        repeat(choice(
+          $.attribute_specifier,
+          $.identifier,
+          alias($.preproc_call_expression, $.call_expression),
+        )),
+      ),
+    ),
+
+    _function_declaration_declarator: $ => prec.right(1,
       seq(
         field('declarator', $._declarator),
         field('parameters', $.parameter_list),
         optional($.gnu_asm_expression),
         repeat($.attribute_specifier),
       )),
+
     function_field_declarator: $ => prec(1, seq(
       field('declarator', $._field_declarator),
       field('parameters', $.parameter_list),
@@ -449,6 +496,11 @@ module.exports = grammar({
       field('parameters', $.parameter_list),
     )),
 
+    _old_style_function_declarator: $ => seq(
+      field('declarator', $._declarator),
+      field('parameters', alias($._old_style_parameter_list, $.parameter_list)),
+    ),
+
     array_declarator: $ => prec(1, seq(
       field('declarator', $._declarator),
       '[',
@@ -525,17 +577,43 @@ module.exports = grammar({
       $._type_identifier,
     ),
 
-    sized_type_specifier: $ => seq(
-      repeat1(choice(
-        'signed',
-        'unsigned',
-        'long',
-        'short',
-      )),
-      field('type', optional(choice(
-        prec.dynamic(-1, $._type_identifier),
-        $.primitive_type,
-      ))),
+    sized_type_specifier: $ => choice(
+      seq(
+        repeat(choice(
+          'signed',
+          'unsigned',
+          'long',
+          'short',
+        )),
+        field('type', optional(choice(
+          prec.dynamic(-1, $._type_identifier),
+          $.primitive_type,
+        ))),
+        repeat1(choice(
+          'signed',
+          'unsigned',
+          'long',
+          'short',
+        )),
+      ),
+      seq(
+        repeat1(choice(
+          'signed',
+          'unsigned',
+          'long',
+          'short',
+        )),
+        field('type', optional(choice(
+          prec.dynamic(-1, $._type_identifier),
+          $.primitive_type,
+        ))),
+        repeat(choice(
+          'signed',
+          'unsigned',
+          'long',
+          'short',
+        )),
+      ),
     ),
 
     primitive_type: _ => token(choice(
@@ -573,8 +651,20 @@ module.exports = grammar({
 
     enumerator_list: $ => seq(
       '{',
-      commaSep($.enumerator),
-      optional(','),
+      repeat(choice(
+        seq($.enumerator, ','),
+        alias($.preproc_if_in_enumerator_list, $.preproc_if),
+        alias($.preproc_ifdef_in_enumerator_list, $.preproc_ifdef),
+        seq($.preproc_call, ','),
+      )),
+      optional(seq(
+        choice(
+          $.enumerator,
+          alias($.preproc_if_in_enumerator_list_no_comma, $.preproc_if),
+          alias($.preproc_ifdef_in_enumerator_list_no_comma, $.preproc_ifdef),
+          $.preproc_call,
+        ),
+      )),
       '}',
     ),
 
@@ -622,13 +712,14 @@ module.exports = grammar({
 
     field_declaration: $ => seq(
       $._declaration_specifiers,
-      commaSep(seq(
-        field('declarator', $._field_declarator),
-        optional($.bitfield_clause),
-      )),
+      optional($._field_declaration_declarator),
       optional($.attribute_specifier),
       ';',
     ),
+    _field_declaration_declarator: $ => commaSep1(seq(
+      field('declarator', $._field_declarator),
+      optional($.bitfield_clause),
+    )),
 
     bitfield_clause: $ => seq(':', $._expression),
 
@@ -637,15 +728,18 @@ module.exports = grammar({
       optional(seq('=', field('value', $._expression))),
     ),
 
-    variadic_parameter: _ => seq(
-      '...',
-    ),
+    variadic_parameter: _ => '...',
 
     parameter_list: $ => seq(
       '(',
       commaSep(choice($.parameter_declaration, $.variadic_parameter)),
       ')',
     ),
+    _old_style_parameter_list: $ => seq(
+      '(',
+      commaSep(choice($.identifier, $.variadic_parameter)),
+      ')',
+    ),
 
     parameter_declaration: $ => seq(
       $._declaration_specifiers,
@@ -681,6 +775,8 @@ module.exports = grammar({
       $.break_statement,
       $.continue_statement,
       $.goto_statement,
+      $.seh_try_statement,
+      $.seh_leave_statement,
     ),
 
     _top_level_statement: $ => choice(
@@ -765,14 +861,18 @@ module.exports = grammar({
     for_statement: $ => seq(
       'for',
       '(',
+      $._for_statement_body,
+      ')',
+      field('body', $._statement),
+    ),
+    _for_statement_body: $ => seq(
       choice(
         field('initializer', $.declaration),
         seq(field('initializer', optional(choice($._expression, $.comma_expression))), ';'),
       ),
-      field('condition', optional(choice($._expression, $.comma_expression))), ';',
+      field('condition', optional(choice($._expression, $.comma_expression))),
+      ';',
       field('update', optional(choice($._expression, $.comma_expression))),
-      ')',
-      field('body', $._statement),
     ),
 
     return_statement: $ => seq(
@@ -795,6 +895,27 @@ module.exports = grammar({
       ';',
     ),
 
+    seh_try_statement: $ => seq(
+      '__try',
+      field('body', $.compound_statement),
+      choice($.seh_except_clause, $.seh_finally_clause),
+    ),
+
+    seh_except_clause: $ => seq(
+      '__except',
+      field('filter', $.parenthesized_expression),
+      field('body', $.compound_statement),
+    ),
+
+    seh_finally_clause: $ => seq(
+      '__finally',
+      field('body', $.compound_statement),
+    ),
+
+    seh_leave_statement: _ => seq(
+      '__leave', ';',
+    ),
+
     // Expressions
 
     _expression: $ => choice(
@@ -819,16 +940,20 @@ module.exports = grammar({
       $.compound_literal_expression,
       $.identifier,
       $.number_literal,
-      $.string_literal,
+      $._string,
       $.true,
       $.false,
       $.null,
-      $.concatenated_string,
       $.char_literal,
       $.parenthesized_expression,
       $.gnu_asm_expression,
     ),
 
+    _string: $ => prec.left(choice(
+      $.string_literal,
+      $.concatenated_string,
+    )),
+
     comma_expression: $ => seq(
       field('left', $._expression),
       ',',
@@ -838,7 +963,7 @@ module.exports = grammar({
     conditional_expression: $ => prec.right(PREC.CONDITIONAL, seq(
       field('condition', $._expression),
       '?',
-      optional(field('consequence', $._expression)),
+      optional(field('consequence', choice($._expression, $.comma_expression))),
       ':',
       field('alternative', $._expression),
     )),
@@ -978,7 +1103,7 @@ module.exports = grammar({
       choice('asm', '__asm__'),
       repeat($.gnu_asm_qualifier),
       '(',
-      field('assembly_code', choice($.string_literal, $.concatenated_string)),
+      field('assembly_code', $._string),
       optional(seq(
         field('output_operands', $.gnu_asm_output_operand_list),
         optional(seq(
@@ -1034,7 +1159,7 @@ module.exports = grammar({
 
     gnu_asm_clobber_list: $ => seq(
       ':',
-      commaSep(field('register', $.string_literal)),
+      commaSep(field('register', $._string)),
     ),
 
     gnu_asm_goto_list: $ => seq(
@@ -1077,14 +1202,27 @@ module.exports = grammar({
       '}',
     ),
 
-    initializer_pair: $ => seq(
-      field('designator', repeat1(choice($.subscript_designator, $.field_designator))),
-      '=',
-      field('value', choice($._expression, $.initializer_list)),
+    initializer_pair: $ => choice(
+      seq(
+        field('designator', repeat1(choice(
+          $.subscript_designator,
+          $.field_designator,
+          $.subscript_range_designator,
+        ))),
+        '=',
+        field('value', choice($._expression, $.initializer_list)),
+      ),
+      seq(
+        field('designator', $._field_identifier),
+        ':',
+        field('value', choice($._expression, $.initializer_list)),
+      ),
     ),
 
     subscript_designator: $ => seq('[', $._expression, ']'),
 
+    subscript_range_designator: $ => seq('[', field('start', $._expression), '...', field('end', $._expression), ']'),
+
     field_designator: $ => seq('.', $._field_identifier),
 
     number_literal: _ => {
@@ -1095,13 +1233,13 @@ module.exports = grammar({
       const decimalDigits = seq(repeat1(decimal), repeat(seq(separator, repeat1(decimal))));
       return token(seq(
         optional(/[-\+]/),
-        optional(choice('0x', '0b')),
+        optional(choice(/0[xX]/, /0[bB]/)),
         choice(
           seq(
             choice(
               decimalDigits,
-              seq('0b', decimalDigits),
-              seq('0x', hexDigits),
+              seq(/0[bB]/, decimalDigits),
+              seq(/0[xX]/, hexDigits),
             ),
             optional(seq('.', optional(hexDigits))),
           ),
@@ -1114,24 +1252,29 @@ module.exports = grammar({
             hexDigits,
           )),
         )),
-        repeat(choice('u', 'l', 'U', 'L', 'f', 'F')),
+        /[uUlLwWfFbBdD]*/,
       ));
     },
 
     char_literal: $ => seq(
       choice('L\'', 'u\'', 'U\'', 'u8\'', '\''),
-      choice(
+      repeat1(choice(
         $.escape_sequence,
         alias(token.immediate(/[^\n']/), $.character),
-      ),
+      )),
       '\'',
     ),
 
-    concatenated_string: $ => seq(
-      choice($.identifier, $.string_literal),
-      $.string_literal,
-      repeat(choice($.string_literal, $.identifier)), // Identifier is added to parse macros that are strings, like PRIu64
-    ),
+    // Must concatenate at least 2 nodes, one of which must be a string_literal.
+    // Identifier is added to parse macros that are strings, like PRIu64.
+    concatenated_string: $ => prec.right(seq(
+      choice(
+        seq($.identifier, $.string_literal),
+        seq($.string_literal, $.string_literal),
+        seq($.string_literal, $.identifier),
+      ),
+      repeat(choice($.string_literal, $.identifier)),
+    )),
 
     string_literal: $ => seq(
       choice('L"', 'u"', 'U"', 'u8"', '"'),
@@ -1164,7 +1307,8 @@ module.exports = grammar({
     null: _ => choice('NULL', 'nullptr'),
 
     identifier: _ =>
-      /(\p{XID_Start}|_|\\u[0-9A-Fa-f]{4}|\\U[0-9A-Fa-f]{8})(\p{XID_Continue}|\\u[0-9A-Fa-f]{4}|\\U[0-9A-Fa-f]{8})*/,
+      // eslint-disable-next-line max-len
+      /(\p{XID_Start}|\$|_|\\u[0-9A-Fa-f]{4}|\\U[0-9A-Fa-f]{8})(\p{XID_Continue}|\$|\\u[0-9A-Fa-f]{4}|\\U[0-9A-Fa-f]{8})*/,
 
     _type_identifier: $ => alias(
       $.identifier,
@@ -1215,9 +1359,11 @@ module.exports.PREC = PREC;
  *
  * @param {RuleBuilder<string>} content
  *
+ * @param {number} precedence
+ *
  * @return {RuleBuilders<string, string>}
  */
-function preprocIf(suffix, content) {
+function preprocIf(suffix, content, precedence = 0) {
   /**
     *
     * @param {GrammarSymbols<string>} $
@@ -1225,50 +1371,51 @@ function preprocIf(suffix, content) {
     * @return {ChoiceRule}
     *
     */
-  function elseBlock($) {
+  function alternativeBlock($) {
     return choice(
       suffix ? alias($['preproc_else' + suffix], $.preproc_else) : $.preproc_else,
       suffix ? alias($['preproc_elif' + suffix], $.preproc_elif) : $.preproc_elif,
+      suffix ? alias($['preproc_elifdef' + suffix], $.preproc_elifdef) : $.preproc_elifdef,
     );
   }
 
   return {
-    ['preproc_if' + suffix]: $ => seq(
+    ['preproc_if' + suffix]: $ => prec(precedence, seq(
       preprocessor('if'),
       field('condition', $._preproc_expression),
       '\n',
       repeat(content($)),
-      field('alternative', optional(elseBlock($))),
+      field('alternative', optional(alternativeBlock($))),
       preprocessor('endif'),
-    ),
+    )),
 
-    ['preproc_ifdef' + suffix]: $ => seq(
+    ['preproc_ifdef' + suffix]: $ => prec(precedence, seq(
       choice(preprocessor('ifdef'), preprocessor('ifndef')),
       field('name', $.identifier),
       repeat(content($)),
-      field('alternative', optional(choice(elseBlock($), $.preproc_elifdef))),
+      field('alternative', optional(alternativeBlock($))),
       preprocessor('endif'),
-    ),
+    )),
 
-    ['preproc_else' + suffix]: $ => seq(
+    ['preproc_else' + suffix]: $ => prec(precedence, seq(
       preprocessor('else'),
       repeat(content($)),
-    ),
+    )),
 
-    ['preproc_elif' + suffix]: $ => seq(
+    ['preproc_elif' + suffix]: $ => prec(precedence, seq(
       preprocessor('elif'),
       field('condition', $._preproc_expression),
       '\n',
       repeat(content($)),
-      field('alternative', optional(elseBlock($))),
-    ),
+      field('alternative', optional(alternativeBlock($))),
+    )),
 
-    ['preproc_elifdef' + suffix]: $ => seq(
+    ['preproc_elifdef' + suffix]: $ => prec(precedence, seq(
       choice(preprocessor('elifdef'), preprocessor('elifndef')),
       field('name', $.identifier),
       repeat(content($)),
-      field('alternative', optional(elseBlock($))),
-    ),
+      field('alternative', optional(alternativeBlock($))),
+    )),
   };
 }
 
